#include <stdio.h>

#include <dslock.h>
#include <dserror.h>
#include <opcserver.h>
#include <opcserverinterface.h>

#ifdef _WIN32
# define OPC_LABVIEW_DATA_DIRECTORY "C:\\OPC"
# define export __declspec(dllexport)
#else
# define export
# define OPC_LABVIEW_DATA_DIRECTORY NULL
#endif /* _WIN32 */

static OPCServerContext ctx = NULL;
static OPCInterface iface = NULL;
static OPCFDOPCBinaryDictionaryContext dict;
static DSLockContextS lock; 
static int status = 0;


void export opcLabviewStop() {
    if (!status) return;
    
    dsLockWrite(&lock);
    if (!ctx) return;

    dsDebugServer(DS_DEBUG_NOTICE, "opcLabview: Stopping Server");
    
    if (iface) opcDestroyInterface(iface);
    opcServerDestroyContext(ctx);
    ctx = NULL;
    dsLockReturn(&lock);

    dsDebugServer(DS_DEBUG_NOTICE, "opcLabview: Server Stopped");
}

int export opcLabviewStart() {
    int err;
    
    if (!status) return -1;
    dsLockWrite(&lock);
    if (ctx) return 0;

    dsDebugServer(DS_DEBUG_NOTICE, "opcLabview: Starting Server");

    ctx = opcServerCreateContext();

    if (ctx) err = opcServerInitContext(ctx, OPC_LABVIEW_DATA_DIRECTORY, OPC_SERVER_NO_BACKEND);
    else err = DS_ERR_MEM;

    if (!err) err = opcServerRun(ctx);
    
    if (!err) {
	iface = opcCreateInterface(ctx->ctx);
	dict = opcFDOPCBinaryCreateDictionary(ctx->ctx->fd_opcbinary, "labview.xml");
	if (dict) opcFDOPCBinaryDictionaryLocalise(dict);
    }

/* DS: start mantainer thread */
    dsLockReturn(&lock);

    if (err) opcLabviewStop();
    else dsDebugServer(DS_DEBUG_NOTICE, "opcLabview: Server Started");

    return err;
}



#ifdef _WIN32
BOOL WINAPI DllMain(HINSTANCE dllhandle, DWORD reason, LPVOID reserved)
{
    int err;
    switch (reason) {
	case DLL_PROCESS_ATTACH:
	break;
	
	case DLL_PROCESS_DETACH:
	break;

	case DLL_THREAD_ATTACH:
        break;

	case DLL_THREAD_DETACH:
    	break;
    }

    return TRUE;
}
#else
void _init() {
    int err;
    ctx = NULL;
    dsDebugSetMethod(DS_DEBUG_METHOD_FILE, OPC_DIR "/opc.log");
    dsDebugSetLevel(DS_DEBUG_DEBUG);
    dsDebugServer(DS_DEBUG_NOTICE, "opcLabview: Loading");

    err = opcServerInit();
    if (!err) err = dsLockInitContext(&lock);
    if (!err) status = 1;

    dsDebugServer(DS_DEBUG_WARNING, "opcLabview: Loaded");
}

void _fini() {
    if (!status) return;
    dsDebugServer(DS_DEBUG_WARNING, "opcLabview: Unloading");
    opcLabviewStop();
    dsLockFreeContext(&lock);
    opcServerDestroy();
    dsDebugServer(DS_DEBUG_NOTICE, "opcLabview: Unloaded");
}
#endif /* _WIN32 */



int export opcLabviewRegisterItem() {
    return 0;    
}

int export opcLabviewPublish(const char *path, const char *name, void *value) {
    int err;
    OPCItemContext ictx;
    OPCType type;
    OPCDataContext dm;
    
    if (!status) return -1;
    
    dsLockWrite(&lock);
    if (ctx) {
//	dsDebugServer(DS_DEBUG_NOTICE, "opcLabview: %s/%s = %lf", path, name, *value);
	
	ictx = opcInterfaceGetItem(iface, path, name);
	if (ictx) {
	    dm = (OPCDataContext)ictx->dm;
	    err = 0;
	} else {
	    err = opcServerRegisterItem(ctx->ctx, &ictx, path, name);
	    if (err) ictx = NULL;
	    else {
		type = opcFDOPCBinaryDictionaryCreateType(dict, name);
		if (type) {
		    dm = opcItemRegisterLegacyDataManager(ictx, type, NULL, OPC_DATA_FLAG_SEQUENTIAL);
		    opcItemEnableDataManager(ictx);
		}
		opcItemResetManualMode(ictx);
	    }
	    
	    err = 1;
	}


	if (ictx) {
	    opcDataPushData(dm, OPC_REPRESENTATION_NATIVE, NULL, value, 0);
	    if (!err) opcItemReturn(ictx);
	}
	
    }
    dsLockReturn(&lock);
    return 0;
}



/*
int __declspec(dllexport) opcLabviewPushData(const char *ItemName, const char *DataType, const void *data, unsigned long options, unsigned long timeout) {
}

/*
opcLabviewPushData
opcLabviewWaitRequest

opcLabviewReadData
//opcLabviewWaitData

Properties

Labview Client
*/
