#include <opcassert.h>
#include <opcdebug.h>

#include <opcdatarepresentation.h>

static inline void opcDataRepresentationZeroContext(OPCDataRepresentationContext ctx) {
    assert(ctx);
    ctx->allocator = NULL;
    ctx->format = NULL;
    ctx->id = 0;
    ctx->base = NULL;
}

OPCDataRepresentationContext opcDataRepresentationCreateContext(OPCMallocContext mctx) {
    OPCDataRepresentationContext ctx;

    ctx = (OPCDataRepresentationContext)opcMalloc(mctx,sizeof(struct OPCDataRepresentationContextT));
    if (ctx) {
	ctx->malloc=mctx;
	opcDataRepresentationZeroContext(ctx);
    }
    return ctx;
}

int opcDataRepresentationInitContext(OPCDataRepresentationContext ctx, OPCFormatContext format, OPCType type, OPCAllocatorContext allocator) {
    opcDataRepresentationZeroContext(ctx);
    
    ctx->type = type;
    ctx->format = format;
    ctx->allocator = allocator;
    
    return 0;
}

int  opcDataRepresentationSetId(OPCDataRepresentationContext ctx, unsigned short int id) {
    assert(ctx);
    
    ctx->id = id;
    return 0;
}

int opcDataRepresentationSetBaseRepresentation(OPCDataRepresentationContext ctx, OPCDataRepresentationContext base) {
    assert(ctx);

    ctx->base = base;
    return 0;
}

void opcDataRepresentationFreeContext(OPCDataRepresentationContext ctx) {
    assert(ctx);
    
    if (ctx->allocator) opcAllocatorDestroyContext(ctx->allocator);
    opcDataRepresentationZeroContext(ctx);
}

void opcDataRepresentationDestroyContext(OPCDataRepresentationContext ctx) {
    opcDataRepresentationFreeContext(ctx);
    opcFree(ctx->malloc, ctx);
}


int opcDataRepresentationImportData(OPCDataRepresentationContext ctx, OPCDataChunkContext chunk, const void *data, unsigned long size) {
    assert(ctx);

    return opcFormatImportData(ctx->format, ctx->type, ctx->allocator, opcDataChunkGetData(chunk, ctx->id), data, size);
}

int opcDataRepresentationGet(OPCDataRepresentationContext ctx, OPCDataChunkContext chunk) {
    int err = 0;
    OPCData data, native;
    unsigned short available;
    unsigned long representations;    

    assert(ctx);
    assert(chunk);


    opcDataChunkLock(chunk);
    data=opcDataChunkGet(chunk, ctx->id);
    if (!data) {
	available = opcDataChunkFindData(chunk);
	if (available!=(unsigned short)-1) {
	    assert(ctx->base);
	    native=opcDataChunkGetData(chunk, ctx->base->id);

	    if (available) {
		opcDebugServer(OPC_DEBUG_DEBUG, "opcDataRepresentation: Convert %u->0",available);
		err = opcFormatWriteNative(ctx->format, ctx->type, ctx->base->allocator, native, opcDataChunkGetData(chunk, available));
	    } 
	    if ((!err)&&(ctx->id)) {
		opcDebugServer(OPC_DEBUG_DEBUG, "opcDataRepresentation: Convert 0->%u",ctx->id);
		err = opcFormatReadNative(ctx->format, ctx->type, ctx->allocator, opcDataChunkGetData(chunk, ctx->id), native);
	    }
	}
    } else err = 0;
    opcDataChunkUnlock(chunk);

    return err;
}


void opcDataRepresentationClear(OPCDataRepresentationContext ctx, OPCData data) {
    opcFormatFreeData(ctx->format, ctx->type, ctx->allocator, data);
}
