#include <assert.h>

#include <opcerr.h>

#include <opcformatbinary.h>

int opcFormatBinaryIntegerRead(OPCFormatBinaryContext ctx, OPCFormatBinaryIntegerConfig config, void *buffer, void *integer) {
    unsigned char i, bytes;
    OPCBinaryConfigurationContext system;

    assert(ctx);
    assert(config);
    system = ((OPCFormatContext)ctx)->system;
    assert(system);
    
    if ((config->bytes>1)&&(config->byte_order!=system->endianess)) {
	bytes = config->bytes-1;
	for (i=0;i<=bytes;i++)
	    ((char*)integer)[i]=((char*)buffer)[bytes-i];
    } else memcpy(integer,buffer,config->bytes);
    return 0;
}

int opcFormatBinaryIntegerWrite(OPCFormatBinaryContext ctx, OPCFormatBinaryIntegerConfig config, void *buffer, void *integer) {
    return opcFormatBinaryIntegerRead(ctx,config,integer,buffer);
}


/*
int opcFormatBinaryIntegerSize(OPCFormatBinaryContext ctx, OPCFormatBinaryIntegerConfig config) {
}
*/

int opcFormatBinaryFloatRead(OPCFormatBinaryContext ctx, OPCFormatBinaryFloatConfig config, void *buffer, void *number) {
    unsigned char i, bytes;
    OPCBinaryConfigurationContext system;

    assert(ctx);
    assert(config);
    system = ((OPCFormatContext)ctx)->system;
    assert(system);
    
    if (config->float_format != system->float_format) {
	// ToDo: Somehow convert formats
	return ERR_OPC_ITEM_DATA_CONVERSION_FLOAT;
    }
    
    if ((config->bytes!=system->float_size)&&(config->bytes!=system->double_size)&&(config->bytes==system->long_double_size)) {
	// ToDo: Somehow reduce resize floats
	return ERR_OPC_ITEM_DATA_CONVERSION_FLOAT;
    } 
    

    if (config->byte_order!=system->float_endianess) {
	    if ((config->byte_order==MIXED_ENDIAN)||(system->endianess==MIXED_ENDIAN)) {
		// ToDo: convert MIXED Endian
		return ERR_OPC_ITEM_DATA_CONVERSION_FLOAT;
	    } else {
		bytes = config->bytes-1;
		for (i=0;i<=bytes;i++)
		    ((char*)number)[i]=((char*)buffer)[bytes-i];
	    }
    } else memcpy(number,buffer,config->bytes);
    
    return 0;

}

int opcFormatBinaryFloatWrite(OPCFormatBinaryContext ctx, OPCFormatBinaryFloatConfig config, void *buffer, void *number) {
    unsigned char i, bytes;
    OPCBinaryConfigurationContext system;

    assert(ctx);
    assert(config);
    system = ((OPCFormatContext)ctx)->system;
    assert(system);
    
    if (config->float_format != system->float_format) {
	// ToDo: Somehow convert formats
	return ERR_OPC_ITEM_DATA_CONVERSION_FLOAT;
    }
    
    if ((config->bytes!=system->float_size)&&(config->bytes!=system->double_size)&&(config->bytes==system->long_double_size)) {
	// ToDo: Somehow reduce resize floats
	return ERR_OPC_ITEM_DATA_CONVERSION_FLOAT;
    } 
    

    if (config->byte_order!=system->float_endianess) {
	    if ((config->byte_order==MIXED_ENDIAN)||(system->endianess==MIXED_ENDIAN)) {
		// ToDo: convert MIXED Endian
		return ERR_OPC_ITEM_DATA_CONVERSION_FLOAT;
	    } else {
		bytes = config->bytes-1;
		for (i=0;i<=bytes;i++)
		    ((char*)buffer)[i]=((char*)number)[bytes-i];
	    }
    } else memcpy(buffer,number,config->bytes);
    
    return 0;
}


int opcFormatBinaryStructRead(OPCFormatBinaryContext ctx, OPCFormatBinaryStructConfig config, void *buffer, void *irep) {
}

int opcFormatBinaryStructWrite(OPCFormatBinaryContext ctx, OPCFormatBinaryStructConfig config, void *buffer, void *irep) {
}



/*

opcFormatBinaryWriteInteger(OPCBinaryFormatContext ctx,


opcFormatBinaryReadFloat
opcFormatBinaryWriteFloat


opcFormatBinaryReadPointer
opcFormatBinaryWritePointer
opcFormatBinaryReadDate
opcFormatBinaryWriteDate




opcFormatBinaryReadString(

opcFormatBinaryWriteString(

opcFormatBinaryReadStruct(
opcFormatBinaryWriteStruct(
*/