#include <stdio.h>
#include <assert.h>

#define _OPC_PROPERTY_C
#include <opcerr.h>
#include <opcdebug.h>

#include <opcmalloc.h>

#include <opc.h>
#include <opcproperty.h>


OPCPropertyContext opcPropertyCreateContext() {
    unsigned long i;
    OPCPropertyContext ctx;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcPropertyCreateContext");    

    ctx = (OPCPropertyContext)opcMalloc(NULL,sizeof(struct OPCPropertyContextT));
    if (ctx) {
	ctx->id=NULL;
	ctx->instances=NULL;
	for (i=0;i<opc_properties_maxid;i++) ctx->prop[i]=NULL;
    }
    return ctx;
}

int opcPropertyInitContext(OPCPropertyContext ctx) {
    unsigned long i;
    int err=0;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcPropertyInitContext");    

    assert(ctx);
    
    ctx->instances=opcStringListCreate(NULL);
    if (!ctx->instances) return ERR_OPC_MEM;

    ctx->id=HashTableInit(NULL,opc_properties_maxid,HASH_TYPE_POINTER,0,NULL);
    if (!ctx->id) {
	opcPropertyFreeContext(ctx);
	return ERR_OPC_PROPERTIES_INIT;
    }

    for (i=0;opc_property[i].id;i++) {
	err+=HashTableInsert(ctx->id,(void*)&opc_property[i]);
	ctx->prop[opc_property[i].id]=&(opc_property[i]);
    }

    if (err) {
	opcPropertyFreeContext(ctx);
	return ERR_OPC_PROPERTIES_INIT;
    }

    return 0;
}

void opcPropertyFreeContext(OPCPropertyContext ctx) {
    unsigned long i;
    OPCStringList sltmp;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcPropertyFreeContext");    

    assert(ctx);

    if (ctx->id) {
	HashTableDestroy(ctx->id);
	ctx->id=NULL;
    }

    if (ctx->instances) {
	for (sltmp=ctx->instances;sltmp;sltmp=sltmp->next) {
	    if (sltmp->item) opcPropertyReturnInstance(*(OPCPropertyInstanceContext*)(sltmp->item));
	}
	opcStringListDestroy(ctx->instances);
	ctx->instances=NULL;
    }

    for (i=0;i<opc_properties_maxid;i++) ctx->prop[i]=NULL;
}

void opcPropertyDestroyContext(OPCPropertyContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcPropertyDestroyContext");    

    opcPropertyFreeContext(ctx);
    opcFree(NULL,ctx);
}

OPCPropertyInstanceContext opcPropertyGetInstance(OPCPropertyContext ctx) {
    OPCPropertyInstanceContext ictx;
    unsigned long i;

    assert(ctx);
    assert(ctx->instances);
    
    ictx = (OPCPropertyInstanceContext)opcMalloc(NULL, sizeof(struct OPCPropertyInstanceContextT));
    if (!ictx) return NULL;
    
    ictx->pctx=ctx;
    for (i=0;i<opc_properties_maxid;i++) ictx->item[i]=NULL;
    
    if (opcStringListAdd(ctx->instances,&ictx,sizeof(OPCPropertyInstanceContext*))) {
	opcFree(NULL,ictx);
	return NULL;
    }

    return ictx;
}

void opcPropertyFreeInstance(OPCPropertyInstanceContext ctx) {
    unsigned long i;

    assert(ctx);

    for (i=0;i<opc_properties_maxid;i++) {
	if (ctx->item[i]) {
	    opcFree(NULL,ctx->item[i]);
	    ctx->item[i]=NULL;
	}
    }
}

void opcPropertyReturnInstance(OPCPropertyInstanceContext ctx) {
    opcPropertyFreeInstance(ctx);
    opcFree(NULL,ctx);
}

int opcPropertyBinarySetById(OPCPropertyInstanceContext ctx, unsigned long id, void *value, unsigned long size, unsigned char gift) {
    OPCProperty prop;
    
    assert(ctx);
    assert(ctx->pctx);
    assert(value);
    assert(id<opc_properties_maxid);
    assert(size>0);
    
    prop=ctx->pctx->prop[id];
    
    if (ctx->item[id]) {
	if ((prop)&&(prop->type==OPC_TYPE_COMPLEX)) {
	    // Call free function
	} else {
	    opcFree(NULL,ctx->item[id]);
	}
    }
    
    if (gift) ctx->item[id]=value;
    else {
	if ((prop)&&(prop->type==OPC_TYPE_COMPLEX)) {
	    // Call copy function
	} else {
	    ctx->item[id]=(void*)opcMalloc(NULL,size);
	    if (!ctx->item[id]) return ERR_OPC_MEM;
	    memcpy(ctx->item[id],value,size);
	}
    }
    
    return 0;
}

int opcPropertyBinarySetByName(OPCPropertyInstanceContext ctx, const char *name, void *value, unsigned long size, unsigned char gift) {
    OPCProperty prop;
    unsigned long id;
    
    assert(ctx);
    assert(ctx->pctx);
    assert(name);
    
    prop=HashTableFind(ctx->pctx->id,name);
    if (!prop) return ERR_OPC_PROPERTIES_NO_PROPERTY;
    return opcPropertyBinarySetById(ctx,prop->id,value,size,gift);
}

void *opcPropertyBinaryGetById(OPCPropertyInstanceContext ctx, unsigned long id, unsigned char owncopy) {
    assert(ctx);
    assert(id<opc_properties_maxid);
    if (owncopy) {
	// Insert code
    }
    return &(ctx->item[id]);
}

void *opcPropertyBinaryGetByName(OPCPropertyInstanceContext ctx, const char *name, unsigned char owncopy) {
    OPCProperty prop;
    unsigned long id;
    
    assert(ctx);
    assert(ctx->pctx);
    assert(name);
    
    prop=HashTableFind(ctx->pctx->id,name);
    if (!prop) return NULL;
    return opcPropertyBinaryGetById(ctx,prop->id,owncopy);
}

static char *opcPropertySerialize(OPCSimpleType type, const void *value) {
    unsigned long size;
    char *ls;
    assert(value);
    assert(size);

    switch (type) {
	case OPC_TYPE_STRINGLIST:
	break;
	case OPC_TYPE_ENUMERATION:
	break;
	case OPC_TYPE_DOUBLE:
	    ls = (char*)opcMalloc(NULL,32*sizeof(char));
	    if (!ls) return NULL;
	    sprintf(ls, "%le", *(double*)value);
	break;
	case OPC_TYPE_UNSIGNED:
	    ls = (char*)opcMalloc(NULL,32*sizeof(char));
	    if (!ls) return NULL;
	    sprintf(ls, "%lu", *(unsigned long*)value);
	break;	
	case OPC_TYPE_INT:
	case OPC_TYPE_BOOLEAN:
	    ls = (char*)opcMalloc(NULL,32*sizeof(char));
	    if (!ls) return NULL;
	    sprintf(ls, "%li", *(long*)value);
	break;
	case OPC_TYPE_COMPLEX:
	    // Call serealization function
	break;
	default:
	    size = strlen((char*)value) + 1;
	    ls = (char*)opcMalloc(NULL,size*sizeof(char));
	    if (!ls) return NULL;
	    memcpy(ls,value,size);
    }
    return ls;
}

static int opcPropertyUnserialize(OPCSimpleType type, const char *value, void **res, unsigned long *size) {
    unsigned long lsize;
    void *lres;
    
    assert(value);
    assert(res);
    assert(size);

    switch (type) {
	case OPC_TYPE_STRINGLIST:
	break;
	case OPC_TYPE_ENUMERATION:
	break;
	case OPC_TYPE_DOUBLE:
	    lsize = sizeof(double);
	    lres = (void*)opcMalloc(NULL,lsize);
	    if (!lres) return ERR_OPC_MEM;
	    *(double*)lres=atof(value);
	break;
	case OPC_TYPE_UNSIGNED:
	case OPC_TYPE_INT:
	case OPC_TYPE_BOOLEAN:
	    lsize = sizeof(long);
	    lres = (void*)opcMalloc(NULL,lsize);
	    if (!lres) return ERR_OPC_MEM;
	    *(double*)lres=atol(value);
	break;
	case OPC_TYPE_COMPLEX:
	    // Call deserealization function
	break;
	default:
	    lsize = strlen(value)+1;
	    lres =  (void*)opcMalloc(NULL,*size);
	    if (!lres) return ERR_OPC_MEM;
	    memcpy(lres,(void*)value,lsize);
    }
    *res = lres;
    *size = lsize;
    return 0;
}

int opcPropertySetById(OPCPropertyInstanceContext ctx, unsigned long id, const char *value) {
    int err;
    OPCSimpleType type;
    OPCProperty prop;
    unsigned long size;
    void *us;
    
    assert(ctx);
    assert(ctx->pctx);
    assert(id<opc_properties_maxid);
    
    prop=ctx->pctx->prop[id];
    if (prop) type=prop->type;
    else type=OPC_TYPE_STRING;
    
    err = opcPropertyUnserialize(type,value,&us,&size);
    if (err) return err;
    
    return opcPropertyBinarySetById(ctx,id,us,size,1);
}

int opcPropertySetByName(OPCPropertyInstanceContext ctx, const char *name, const char *value) {
    int err;
    OPCProperty prop;
    unsigned long size;
    void *us;
    
    assert(ctx);
    assert(ctx->pctx);

    prop=HashTableFind(ctx->pctx->id,name);
    if (!prop) return ERR_OPC_PROPERTIES_NO_PROPERTY;

    err = opcPropertyUnserialize(prop->type,value,&us,&size);
    if (err) return err;

    return opcPropertyBinarySetById(ctx,prop->id,us,size,1);
}

char *opcPropertyGetById(OPCPropertyInstanceContext ctx, unsigned long id) {
    OPCSimpleType type;
    OPCProperty prop;
    char *s;
    
    assert(ctx);
    assert(ctx->pctx);
    assert(id<opc_properties_maxid);
    
    prop=ctx->pctx->prop[id];
    if (prop) type=prop->type;
    else type=OPC_TYPE_STRING;
    
    return opcPropertySerialize(type,opcPropertyBinaryGetById(ctx,id,0));
}

char *opcPropertyGetByName(OPCPropertyInstanceContext ctx, const char *name) {
    OPCSimpleType type;
    OPCProperty prop;
    char *s;
    
    assert(ctx);
    assert(ctx->pctx);

    prop=HashTableFind(ctx->pctx->id,name);
    if (!prop) return NULL;
    
    return opcPropertySerialize(prop->type,opcPropertyBinaryGetById(ctx,prop->id,0));
}

int opcPropertyInit(OPCContext ctx) {
    int err;
    OPCPropertyContext pctx;

    assert(ctx);

    pctx=opcPropertyCreateContext();
    if (!pctx) return ERR_OPC_MEM;

    err=opcPropertyInitContext(pctx);
    if (err) opcFree(NULL,pctx);
    else ctx->pctx=pctx;
    return err;
}
