#include <stdio.h>

#include <opclib.h>
#include <dsassert.h>
#include <dsplugin.h>
#include <dsdebug.h>
#include <dsbinaryconfiguration.h>


#include <opc.h>
#include <opcinterface.h>
#include <opcserverinterface.h>
#include <opcquality.h>

#include <random.h>



#ifdef LEGACY_DATA_MANAGER
#ifdef PUSH_BENCH
static unsigned long minbench=-1;
static unsigned long maxbench=0;
static double bench=0;
static unsigned long iter = 0;
#endif
#endif


RandomContext randomCreateContext() {
    RandomContext ctx;
    dsDebugServer(DS_DEBUG_NOTICE, "opcRandomCreateContext");    
    ctx = (RandomContext)malloc(sizeof(struct RandomContextT));
    return ctx;
}

static inline_function int PushData(RandomContext ctx) {
    double *arr;
    unsigned long i;

#ifdef FAST_DATA_MANAGER
    DSAllocation allocation;
    struct _random_complex_t *item;

    allocation=dsAllocatorSimpleAllocate(ctx->allocator, sizeof(struct _random_complex_t));
    item = opcAllocatorGet(ctx->allocator,allocation,NULL);
    arr = item->array;
    for (i=0;i<SIZE;i++) 
	arr[i] = (1.*i/100);

    return opcDataFastPushAllocation(ctx->dm, 0, NULL, allocation);
#endif
#ifdef LEGACY_DATA_MANAGER
#ifdef PUSH_BENCH
    unsigned long k, j;
    DSAllocation allocation;
    struct _random_complex_t *item;
    struct OPCTimeT time1, time2;
    struct OPCDurationT duration;
    unsigned long value;

/*
    allocation = dsAllocatorSimpleAllocate(ctx->allocator, sizeof(struct _random_complex_t));
    item = opcAllocatorGet(ctx->allocator,allocation,NULL);

    arr = item->array;
    for (i=0;i<SIZE;i++) 
	arr[i] = (1.*i/100);
*/
    opcTimeCurrentTime(&time1);
    allocation=dsAllocatorStart(ctx->allocator, 0);
    opcAllocatorAllocate(ctx->allocator, allocation, sizeof(struct _random_simple_t), 0);
    for (j=0, k=0;j<STEPS;j++) {
	arr=dsAllocatorAllocate(ctx->allocator, allocation, sizeof(double)*SIZE/STEPS, opc_binary_configuration->align_double);
	for (i=0;i<SIZE/STEPS;i++,k++) 
	    arr[i] = (1.*k/100);
    }
    dsAllocatorAcceptChunk(ctx->allocator, allocation, &k);
    allocation=dsAllocatorAccept(ctx->allocator, allocation);
    dsTimeCurrentTime(&time2);
    dsTimeCalcDuration(&duration, &time1, &time2);

    value = duration.s*1000000 + duration.us;
    printf("Current: %luus\n", value);
    if (iter>=SKIP) {    
	if (value>maxbench) maxbench=value;
	if (value<minbench) minbench=value;
	bench=(bench*(iter-SKIP)+value)/((iter-SKIP)+1);
        printf("Time   : %.1lf (%lu,%lu)\n", bench, minbench, maxbench);
    }
    iter++;
    
    
    return opcDataPushAllocation(ctx->dm, OPC_REPRESENTATION_NATIVE, ctx->allocator, NULL, allocation);
#else
    struct _random_complex_t item;

    arr = item.array;
    for (i=0;i<SIZE;i++) 
	arr[i] = (1.*i/100);
    return opcDataPushData(ctx->dm, OPC_REPRESENTATION_NATIVE, NULL, &item, sizeof(struct _random_complex_t));
#endif
#endif
}

int randomInitContext(RandomContext ctx, DSPluginContext pctx, OPCContext sctx) {
    int err;
    OPCFDOPCBinaryContext fctx;
//    OPCDataFastContext dm;
//    struct _random_complex_t item = { 65535, 10, 1.84 };
    
    dsDebugSetLevel(DS_DEBUG_DEBUG);
    dsDebugServer(DS_DEBUG_NOTICE, "opcRandomInitContext");    
    assert(sctx);
    fctx = sctx->fd_opcbinary;
    assert(fctx);
    
    err = opcServerRegisterItem(sctx,&ctx->ictx,"/Examples/random/","integer");
    if (ctx->ictx) opcItemResetManualMode(ctx->ictx);
    err = opcServerRegisterItem(sctx,&ctx->ictx,"/Examples/random/","unsigned");
    if (ctx->ictx) opcItemResetManualMode(ctx->ictx);
    err = opcServerRegisterItem(sctx,&ctx->ictx,"/Examples/random/","double");
    if (ctx->ictx) opcItemResetManualMode(ctx->ictx);
    err = opcServerRegisterItem(sctx,&ctx->ictx,"/Examples/random/","complex");

    ctx->dict = opcFDOPCBinaryCreateDictionary(fctx, "random.xml");
    if (!ctx->dict) {
	puts("Error Creating Dict");
	return OPC_ERR_FD_OPCBINARY_DICTIONARY_CREATE;
    }
	
    
    err=opcFDOPCBinaryDictionaryLocalise(ctx->dict);
    if (err) puts("Error Localsing");
    
    ctx->type = opcFDOPCBinaryDictionaryCreateType(ctx->dict, "complex");
    

//    opcItemSetDataType(ctx->ictx, ctx->type);
//    opcItemInitNativeDataContext(ctx->ictx,0,OPCQualityGood,&item);

#ifdef FAST_DATA_MANAGER
    ctx->dm=opcItemRegisterFastDataManager(ctx->ictx, ctx->type, NULL, 0);
#ifdef FAST_DATA_MANAGER_XML
    opcDataFastSetUserizeFunction(ctx->dm, opcUserDataFastGetOPCXML);
#endif
    ctx->allocator=opcDataFastGetAllocator(ctx->dm);
    PushData(ctx);
#endif

#ifdef LEGACY_DATA_MANAGER
#ifdef LEGACY_DATA_MANAGER_DATABUFFER
    ctx->dm=opcItemRegisterLegacyDataManager(ctx->ictx, ctx->type, NULL, OPC_DATA_FLAG_SEQUENTIAL);
#ifdef PUSH_BENCH
    ((OPCDataBufferContext)((OPCDataRepresentationContext)ctx->dm->representation[0])->allocator)->fixed_allocation_size = 0;
#endif
#else
    ctx->dm=opcItemRegisterLegacyDataManager(ctx->ictx, ctx->type, NULL, 0);
#endif 
#ifdef LEGACY_DATA_MANAGER_BINARY
    ctx->ictx->representation = 0;
#endif

#ifdef PUSH_BENCH
    ctx->allocator = opcDataGetAllocator(ctx->dm, OPC_REPRESENTATION_NATIVE, NULL);
#endif
    PushData(ctx);
#endif

    opcItemEnableDataManager(ctx->ictx);
    
    

/*
    opcItemPush(ctx->ictx,time,quality,value); // slow method
    
    opcItemGetNativeFormat(ctx->ictx);
    opcDataNewChunk(...,FORMAT,...);
    opc....Alocate(
    opcDataFinishChunk(...)
    
    =opcItemGetOPCXMLFormat(ctx->ictx);
*/
    if (ctx->ictx) opcItemResetManualMode(ctx->ictx);
    
    ctx->pctx=pctx;
    dsDebugServer(DS_DEBUG_INFO, "opcRandomInitContext Ok");    
    return err;
}

int randomRun(RandomContext ctx) {
    unsigned long i = 0;
    assert(ctx);
    while (dsPluginContinue(ctx->pctx)) {
#ifdef PUSH_BENCH
        PushData(ctx);
#else
	if ((i++)%CLIENTS==0) PushData(ctx);
#endif
#ifdef _WIN32
	Sleep(1000);
#else /* _WIN32 */
	sleep(1);
#endif /* _WIN32 */
    }
    return 0;
}

void randomFreeContext(RandomContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcRandomFreeContext");    
    assert(ctx);
}

void randomDestroyContext(RandomContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcRandomDestroyContext");    
    randomFreeContext(ctx);
    free(ctx);
}

dsPluginInfoFunctionImplementation {
    return &info;
}
