#ifndef _OPC_FORMAT_H
#define _OPC_FORMAT_H

typedef struct OPCFormatContextT *OPCFormatContext;

#include <dslib.h>
#include <dsallocator.h>
#include <opctype.h>
#include <dsbufferpool.h>
#include <opcuserdata.h>

/*
opcFormatReadNative
    Convert Native representation to OPCXML    

    ctx:	OPCXML Format Context
    type: 	Type to read
    allocator:	allocator for OPCXML representation
    data:	data in native format    

opcFormatWriteNative
    Convert OPCXML representation to Native format

    ctx:	OPCXML Format Context
    type: 	Type to read
    allocator:	allocator for Native representation
    data:	data in OPCXML format
*/



typedef enum OPCFormatTypeT OPCFormatType;
typedef enum OPCAllocatorTypeT OPCAllocatorType;

typedef OPCFormatContext (*OPCFormatCreateFunction)(DSMallocContext mctx);
typedef int (*OPCFormatInitFunction)(OPCFormatContext ctx, const void *options);
typedef void (*OPCFormatFreeFunction)(OPCFormatContext ctx);
typedef void (*OPCFormatDestroyFunction)(OPCFormatContext ctx);


typedef int (*OPCFormatImportDataFunction)(OPCFormatContext ctx, OPCType type, DSAllocatorContext allocator, OPCData data, const void *buffer, unsigned long size);
typedef void (*OPCFormatFreeDataFunction)(OPCFormatContext ctx, OPCType type, DSAllocatorContext allocator, OPCData data);
typedef int (*OPCFormatReadNativeFunction)(OPCFormatContext ctx, OPCType type, DSAllocatorContext allocator, OPCData data, OPCData native);
typedef int (*OPCFormatWriteNativeFunction)(OPCFormatContext ctx, OPCType type, DSAllocatorContext allocator, OPCData native, OPCData data);
typedef int (*OPCFormatSendFunction)(OPCFormatContext ctx, OPCType type, DSAllocatorContext allocator, OPCData data, DSSendFunction func, void *attr);


typedef int (*OPCFormatWriteIntegerFunction)(OPCFormatContext ctx, OPCType type, OPCConversion conversion, void *source, unsigned long *bytes);
typedef int (*OPCFormatWriteFloatFunction)(OPCFormatContext ctx, OPCType type, OPCConversion conversion, void *source, unsigned long *bytes);


enum OPCFormatTypeT {
    OPC_FORMAT_TYPE_UNKNOWN = 0,
    OPC_FORMAT_TYPE_NATIVE,
    OPC_FORMAT_TYPE_BINARY,
    OPC_FORMAT_TYPE_OPCXML,
    OPC_FORMAT_TYPE_XML
};

enum OPCAllocatorTypeT {
    OPC_ALLOCATOR_TYPE_ABSENT = 0,
    OPC_ALLOCATOR_TYPE_MINIMAL,
    OPC_ALLOCATOR_TYPE_FULL
};

struct OPCFormatContextT {
    char *title;
    
    OPCFormatType type;
    OPCAllocatorType allocator;

    OPCFormatCreateFunction CreateContext;
    OPCFormatInitFunction InitContext;
    OPCFormatFreeFunction FreeContext;
    OPCFormatDestroyFunction DestroyContext;

    OPCFormatImportDataFunction ImportData;
    OPCFormatFreeDataFunction FreeData;

    OPCFormatReadNativeFunction ReadNative;
    OPCFormatWriteNativeFunction WriteNative;
    OPCFormatSendFunction Send;
    
    OPCFormatWriteIntegerFunction WriteInteger;
    OPCFormatWriteFloatFunction WriteFloat;

    DSMallocContext malloc;
};


OPCFormatContext opcFormatCreateContext(OPCFormatContext format, DSMallocContext mctx);
int opcFormatInitContext(OPCFormatContext ctx, const void *options);
void opcFormatFreeContext(OPCFormatContext ctx);
void opcFormatDestroyContext(OPCFormatContext ctx);

#define opcFormatImportData(ctx,type,allocator,data,buffer,size) \
    (((ctx)&&(ctx->ImportData))?ctx->ImportData(ctx,type,allocator,data,buffer,size):OPC_ERR_FORMAT_MISSING_METHOD_IMPORTDATA)

#define opcFormatReadNative(ctx,type,allocator,data,native) \
    (((ctx)&&(ctx->ReadNative))?ctx->ReadNative(ctx, type, allocator, data, native):OPC_ERR_FORMAT_MISSING_METHOD_READNATIVE)

#define opcFormatWriteNative(ctx,type,allocator,native,data) \
    (((ctx)&&(ctx->WriteNative))?ctx->WriteNative(ctx, type, allocator, native, data):OPC_ERR_FORMAT_MISSING_METHOD_WRITENATIVE)

#define opcFormatFreeData(ctx,type,allocator,data) \
    ((ctx)&&(ctx->FreeData))?ctx->FreeData(ctx, type, allocator, data):0

#define opcFormatSend(ctx,type,allocator,data,func,attr) \
    (((ctx)&&(func))?((ctx->Send)?ctx->Send(ctx,type,allocator,data,func,attr):OPC_ERR_FORMAT_MISSING_METHOD_SEND):DS_ERR_INVALID_ARGUMENT)

/*
#define opcFormatFreeData(ctx,type,allocator,data) \
    if (ctx) { \
	if (ctx->FreeData) ctx->FreeData(ctx, type, allocator, data) \
	else dsAllocatorFree(allocator, data->ptr); \
    }

#define opcFormatSend(ctx,type,allocator,data,func,attr) \
    (((ctx)&&(func))?((ctx->Send)?ctx->Send(ctx,type,allocator,data,func,attr):opcAllocationCall(allocator, (DSAllocation)data->ptr, func, attr, 1)):DS_ERR_INVALID_ARGUMENT)
*/

#endif /* _OPC_FORMAT_H */
