#include <stdio.h>

#include <dserror.h>
#include <dsdebug.h>
#include <dssignal.h>
#include <opcserver.h>

#ifdef _WIN32
#include <windows.h>
static HANDLE hEvent = NULL;
#endif /* _WIN32 */

static int RunFlag = 1;

static void signal_exit_handler(int signo) { 
    RunFlag = 0; 
#ifdef _WIN32 
    if (hEvent) SetEvent(hEvent);
#endif /* _WIN32 */
}

main() {
    int err;
    const char *serr;
    OPCServerContext ctx;
#ifdef _WIN32
    HANDLE hEventCopy;
#endif /* _WIN32 */

    err = dsSignalINTSetAction(signal_exit_handler);
    if (err) {
	printf("Initialization Error: Installing SIGINT handler\n");
	return 0;
    }
    
    err = dsSignalINTBlock();
    if (err) {
	printf("Initialization Error: Blocking SIGINT signal\n");
	return 0;    
    }

    err = opcServerInit();
    if (err) {
	printf("Initialization Error: Server Global Initialization Error: %lx\n", err);
	return 0;
    }
    
    ctx=opcServerCreateContext();
    if (!ctx) {
	serr = dsError(err,NULL);
	opcServerDestroy();
	printf("Initialization Error: Cann't allocate memory for OPCServerContext\n");
	printf("%s\n", serr);
	return 0;
    }
    
    err=opcServerInitContext(ctx, NULL, 0);
    if (err) {
	serr = dsError(err,NULL);
	opcServerDestroyContext(ctx);
	opcServerDestroy();
	printf("Initalization Error: Server Initialization Error: %lx\n", err);
	printf("%s\n", serr);
	return 0;
    }
    
    err=opcServerRun(ctx);
    if (err) {
	serr = dsError(err,NULL);
	opcServerDestroyContext(ctx);
	opcServerDestroy();
	printf("Execution Error: Server Execution Error: %lx\n", err);
	printf("%s\n",serr);
	return 0;
    }
        
#ifdef _WIN32
    hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (!hEvent) {
	err = DS_ERR_WINAPI;
	serr = dsError(err,NULL);
	opcServerDestroyContext(ctx);
	opcServerDestroy();
	printf("Initialization Error: Creating SigINT Event: %lx\n", err);
	printf("%s\n",serr);
	return 0;
    }
#endif /* _WIN32 */

    err = dsSignalINTUnblock();
    if (err) {
	printf("Initialization Error: Unblocking SIGINT signal\n");
    }

    while ((RunFlag)&&(!err)) {
	puts("Iteration");

#ifdef _WIN32
	WaitForSingleObject(hEvent, OPC_MANTAIN_DELAY*1000);
#else /* _WIN32 */
	if (RunFlag) sleep(OPC_MANTAIN_DELAY);
#endif /* _WIN32 */

	opcServerMantain(ctx);
    }

    dsSignalINTBlock();
    if (err) printf("Execution Error: Waiting for SIGINT signal failed!\n");

#ifdef _WIN32
    hEventCopy = hEvent;
    hEvent = NULL;
    CloseHandle(hEventCopy);
#endif /* _WIN32 */

    
    serr = dsError(0,NULL);
    
    opcServerDestroyContext(ctx);
    opcServerDestroy();
    
    printf("\nFinished: %s\n\n",serr);
    return 0;
}
